
/* ----------------------------------------------------------------------
   The SINUS Firewall -- a TCP/IP packet filter for Linux

   Written within the SINUS project at the University of Zurich,
   SWITCH, Telekurs Payserv AG, ETH Zurich.
   originally based on the sf Firewall Software (C) 1996 by Robert
   Muchsel and Roland Schmid.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

   SINUS Firewall resources:
   SINUS Homepage: http://www.ifi.unizh.ch/ikm/SINUS/
   Firewall Homepage: http://www.ifi.unizh.ch/ikm/SINUS/firewall.html
   Frequently asked questions: http://www.ifi.unizh.ch/ikm/SINUS/sf_faq.html
   Mailing list for comments, questions, bug reports: firewall@ifi.unizh.ch
   ----------------------------------------------------------------------  */



package sfclasses;

import java.awt.*;
import java.io.*;
import java.net.*;
import java.util.*;

/**
 * This class implements the communication with the firewall. All methods
 * in this class are static and synchronized.
 * @version 1.0 10 Dec 1996
 * @author Roland E. Schmid
 */

public class Communicator {

  static { // This is a static initializer

  }

  // public methods
	/**
	 * Check if connection exists.
	 * @param conn IP-Address to check for.
	 * @return Connection if a connection to the given IP-Adddress exists,
	 *				 null otherwise.
	 */
	public static Connection findConnection(byte[] conn) {

	Connection c;
	InetAddress iaddr;
	String s;

		s = resolveIP(conn);
		try {
			iaddr = InetAddress.getByName(s);
		}
		catch (UnknownHostException e) {
			System.out.println("Error in Connection.connExists: "+e);
			return null;
		}
	
		Enumeration en = openConns.elements();
		while (en.hasMoreElements()) {
			c = (Connection)en.nextElement();
			if (c.iaddr.equals(iaddr));
				return c;
		}
		return null;
	} // findConnection

  /**
   * Open server connection or return Connection if already open.
   * @param servAddr server address
   * @return Connection serves as interface for communication
   */
  public static synchronized Connection openServer(byte[] server) {

	Connection conn;
	
		conn = findConnection(server);
		if (conn != null) {
			if (!conn.open()) {
				commError = COMM_OPENCONN;
				return null;
			}
			return conn;
		}

		conn = new Connection(server);
		if (!conn.isValid())
			return null;

		if (!conn.open())
			return null;
		openConns.addElement(conn);

		if (!performAuthentication(conn))
			commError = COMM_AUTH_FAILED;
		return conn;
  } // openServer

	public static synchronized boolean performAuthentication(Connection server) {
	byte[] keynum = new byte[4];

		Authentication auth = new Authentication(parent, server.iaddr.getHostName());
		auth.display();
		
		login = auth.getLogin().trim() + "\u0000";
		key = auth.getKey().trim();

		if ((login == null) || (login.length() <= 1)) {
			// user hit Cancel button
			commError = COMM_AUTH_FAILED;
			return false;
		}

		try {
			server.writeHeader(SERV_AUTHENTICATE, key.length() + login.length());
			server.out.writeBytes(key);
			server.out.writeBytes(login);
			server.readHeader();
			switch (server.readCommand) {
				case SERV_SERVER_AUTHOK:
					UserDialog.NoticeBox("Authentification succesfull.");
					commError = COMM_OK;
					server.writeHeader(SERV_CLIENT_AUTHOK, 0);
					return true;
				case SERV_KEY_EXPIRED:
					return performKeyupdate(server);
				case SERV_ERROR_AUTH:
					UserDialog.ErrorBox("Authentification failed.");
					commError = COMM_AUTH_FAILED;
					return false;
				default:
					commError = COMM_ERROR;
					return false;
			}
		}
		catch (Exception e) {
      dispose(server);
      commError = COMM_IO;
      return false;
		}
	} // performAuthentification


	public static synchronized boolean performKeyupdate(Connection conn) {

		Keyupdate ku = new Keyupdate(parent, conn.iaddr.getHostName());
		key = ku.getKey().trim();

		try {
			conn.writeHeader(SERV_KEY_UPDATE, key.length());
			conn.out.writeBytes(key);
			conn.readHeader();
			switch (conn.readCommand) {
				case SERV_OK:
					UserDialog.NoticeBox("Password updated. New Keynumber is 99");
					commError = COMM_OK;
					return true;
				case SERV_ERROR_AUTH:
					commError = COMM_AUTH_FAILED;
					return false;
				default:
					commError = COMM_ERROR;
					return false;
			}
		}
    catch (Exception e) {
      dispose(conn);
      commError = COMM_IO;
      return false;
    }
	} // performKeyupdate


  /**
   * Check generated config file for errors and rename it to firewall.conf
   * @param servAddr Server address
   * @return true if successful, commError is set when returning false
   */
  public static synchronized boolean checkConfig(byte[] servAddr) {

	Connection conn;

		conn = openServer(servAddr);
		if (conn == null)
			return false;

    try {
      conn.writeHeader(SERV_CHECKCONFIG, 0);
      conn.readHeader();
      if (conn.readCommand != SERV_OK) {
        commError = COMM_ERROR;
        conn.in.skipBytes(conn.readLength);
        return false;
      }
      conn.readHeader(); // read result of check operation
      return (conn.readCommand == SERV_OK);
    } // try
    catch (Exception e) {
      dispose(conn);
      commError = COMM_IO;
      return false;
    }
  }

  /**
   * Reconfigure the firewall
   * @param servAddr Server Address
   * @return true if successful, commError is set when returning false
   */
  public static synchronized boolean reconfig(byte[] servAddr) {

	Connection conn;

		conn = openServer(servAddr);
		if (conn == null)
			return false;

    try {
      for (int i = 0; i < 3; i++) { // max. 3 retries when ERROR_NOTCONN
        conn.writeHeader(SERV_RECONFIG, 0);
        conn.readHeader(); // read result of check operation
        if (conn.readCommand == SERV_OK) {
					conn.stop();
					openConns.removeElement(conn);
          try {
            Thread.sleep(10000);
          }
          catch (InterruptedException ignore) {}
          if (openServer(servAddr) == null)
						return false;
					return true;
        }
        else
          return false;
      } // for
      return false;
    } // try
    catch (Exception e) {
      dispose(conn);
      commError = COMM_IO;
      return false;
    }
  }

  /**
   * Fetch file from server and store as temporary file
   * @param servAddr server address
   * @param fname filename
   * @return filename of temporary file, commError is set when returning null
   */
  public static synchronized String getTempFile(byte[] servAddr, String fname) {

	String s;
	byte[] buffer;
	Connection conn;

		conn = openServer(servAddr);
		if (conn == null)
			return null;

		try {

      String filename = fname.trim() + "\u0000";
      int len = filename.length();
      conn.writeHeader(SERV_GET_FILE, len);
      conn.out.writeBytes(filename);
			conn.readHeader();
			switch (conn.readCommand) {
				case SERV_OK:
					break;
				default:
					commError = COMM_IO;
					return null;
			}

			s = new String();
			while (true) {
				conn.readHeader();
				if (conn.readCommand != SERV_FILE_DATA) {
					conn.in.skipBytes(conn.readLength);
					return s;
				}
				else
					if (conn.readLength != 0) {
						buffer = new byte[conn.readLength];
						conn.in.readFully(buffer, 0, conn.readLength);
						s += new String(buffer);
					}
					else return s;
			} // while
		} // try
    catch (Exception e) {
      commError = COMM_IO;
      return null;
    }
  }

  /**
   * Request file from server. This method is called from ServerInputStream
   * @param servAddr server address
   * @param fname filename
   * @return true if successful, commError is set when returning false
   * @see ServerInputStream
   */
  public static boolean requestFile(byte[] servAddr, String fname) {

	Connection conn;

		conn = openServer(servAddr);
		if (conn == null) {
			return false;
			}
	
    try {
      String filename = fname.trim() + "\u0000";
      int len = filename.length();
      conn.writeHeader(SERV_GET_FILE, len);
      conn.out.writeBytes(filename);
			conn.readHeader();
			switch(conn.readCommand) {
				case SERV_OK:
					return true;
				default:
					return false;
			}
    } // try
    catch (Exception e) {
      dispose(conn);
      commError = COMM_IO;
      return false;
    }
  }

  /**
   * Write file to server. This method is called from ServerOutputStream
   * @param servAddr server address
   * @param fname filename
   * @return true if successful, commError is set when returning false
   * @see ServerOutputStream
   */
  public static synchronized boolean writeFile(byte[] servAddr, String fname) {

	Connection conn;

		conn = openServer(servAddr);
		if (conn == null)
			return false;
			
    try {
      for (int i = 0; i < 3; i++) { // max. 3 retries when ERROR_NOTCONN
				String filename = fname.trim() + "\u0000";
				int len = filename.length();
				conn.writeHeader(SERV_WRITE_FILE, len);
				conn.out.writeBytes(filename);
				conn.readHeader();
				switch (conn.readCommand) {
          case SERV_OK:
            return true;
					case SERV_WRITE_FAIL:
						commError = COMM_IO;
						return false;
          default:
            commError = COMM_ERROR;
            conn.in.skipBytes(conn.readLength);
            return false;
        } // switch
      } // for
      dispose(conn);
      commError = COMM_IO;
      return false;
    } // try
    catch (Exception e) {
      dispose(conn);
      commError = COMM_IO;
      return false;
    }
  }

  /**
   * Read active TCP connections from firewall
   * @param host Firewall to read configuration from
   * @return true if successful, commError is set when returning false
   */
  public static synchronized boolean readTcpConns(Host host) {
    int tmpint;
    if (!host.isFirewall) {
      commError = COMM_NOFIREWALL;
      return false;
    }
    byte[] hostAddr = host.HostAddresses.getFirstAddress();
    if (hostAddr == null) {
      commError = COMM_NOADDRESS;
      return false;
    }
		
		Connection conn = openServer(hostAddr);
		if (conn == null)
			return false;

    try {
      conn.writeHeader(SERV_READ_TCP, 0);
      conn.readHeader();
      switch (conn.readCommand) {
        case SERV_OK:
          break;
        default:
          commError = COMM_ERROR;
          conn.in.skipBytes(conn.readLength);
          return false;
      } // switch
      TcpConnection tc;
      host.tcpconns = new Vector(10,10);
      conn.readHeader();
      if (conn.readLength != 4) {
        commError = COMM_ERROR;
        return false;
      }
      tmpint = conn.in.readInt();
      host.actDate = new Date((long)tmpint * 1000);
      conn.readHeader();
      while (conn.readLength == 48) {
        tc = new TcpConnection();
        host.tcpconns.addElement(tc);
        tc.id = conn.in.readInt();
        tc.key = conn.in.readInt();
        tmpint = conn.in.readInt();
        tc.created = new Date((long)tmpint * 1000);
        tc.created_jiffies = conn.in.readInt();
        tmpint = conn.in.readInt();
        tc.lastuse = new Date((long)tmpint * 1000);
        tmpint = conn.in.readInt();
        tc.timeout = new Date((long)tmpint * 1000);
        tc.fromaddr = new byte[4];
        conn.in.readFully(tc.fromaddr, 0, 4);
        tc.fromport = conn.in.readInt();
        tc.toaddr = new byte[4];
        conn.in.readFully(tc.toaddr, 0, 4);
        tc.toport = conn.in.readInt();
        tmpint = conn.in.readInt();
        tc.allpsource = tmpint != 0;
        tc.state = conn.in.readInt();
        conn.readHeader();
      }
      if (conn.readLength != 0) {
        commError = COMM_ERROR;
        return false;
      }
      return true;
    }
    catch (Exception e) {
      dispose(conn);
      commError = COMM_IO;
      return false;
    }
  }

  /**
   * Kill TCP connection on firewall
   * @param servAddr Server address (firewall)
   * @param id TCP connection ID
   * @param key TCP connection key
   * @param created Creation time of TCP connection
   * @return true if successful, commError is set when returning false
   */
  public static synchronized boolean killTcp(byte[] servAddr, 
                                       int id, int key, int created) {

		Connection conn = openServer(servAddr);
		if (conn == null)
			return false;

    try {
      conn.writeHeader(SERV_KILL_TCP, 12);
      conn.out.writeInt(id);
      conn.out.writeInt(key);
      conn.out.writeInt(created);
      conn.readHeader();
      if (conn.readCommand != SERV_OK) {
        commError = COMM_ERROR;
        conn.in.skipBytes(conn.readLength);
        return false;
      }
      return true;
    } // try
    catch (Exception e) {
      dispose(conn);
      commError = COMM_IO;
      return false;
    }
  }

  /**
   * Reset usage counter on firewall
   * @param servAddr Server address (firewall)
   * @return true if successful, commError is set when returning false
   */
  public static synchronized boolean resetUsage(byte[] servAddr) {

		Connection conn = openServer(servAddr);
		if (conn == null)
			return false;

    try {
      conn.writeHeader(SERV_RESET_USAGE, 0);
      conn.readHeader();
      if (conn.readCommand != SERV_OK) {
        commError = COMM_ERROR;
        conn.in.skipBytes(conn.readLength);
        return false;
      }
      return true;
    } // try
    catch (Exception e) {
      dispose(conn);
      commError = COMM_IO;
      return false;
    }
  }

  /**
   * Read active firewall configuration
   * @param host Firewall to read configuration from
   * @return true if successful, commError is set when returning false
   */
  public static synchronized boolean readFwConfig(Host host) {
    int tmpint;
    if (!host.isFirewall) {
      commError = COMM_NOFIREWALL;
      return false;
    }
    byte[] hostAddr = host.HostAddresses.getFirstAddress();
    if (hostAddr == null) {
      commError = COMM_NOADDRESS;
      return false;
    }

		Connection conn = openServer(hostAddr);
		if (conn == null)
			return false;
    try {
      conn.writeHeader(SERV_READ_ADDR, 0);
      conn.readHeader();
      switch (conn.readCommand) {
        case SERV_OK:
          break;
        default:
          commError = COMM_ERROR;
          conn.in.skipBytes(conn.readLength);
          return false;
      } // switch
      conn.readHeader();
      if (conn.readLength != 4) {
        commError = COMM_ERROR;
        return false;
      }
      int numEntries = conn.in.readInt();
      host.addr = new byte[numEntries][4];
      host.mask = new byte[numEntries][4];
      host.port = new short[numEntries];
      host.prend = new short[numEntries];
      int pos = 0;
      conn.readHeader();
      while (conn.readLength == 12) {
        conn.in.readFully(host.addr[pos], 0, 4);
        conn.in.readFully(host.mask[pos], 0, 4);
        host.port[pos] = conn.in.readShort();
        host.prend[pos] = conn.in.readShort();
        conn.readHeader();
        pos++;
      }
      if (conn.readLength != 0) {
        commError = COMM_ERROR;
        return false;
      }

      conn.writeHeader(SERV_READ_RULES, 0);
      conn.readHeader();
      switch (conn.readCommand) {
        case SERV_OK:
          break;
        default:
          commError = COMM_ERROR;
          conn.in.skipBytes(conn.readLength);
          return false;
      } // switch
      host.rules = new Vector(20,20);
      Rule r;
      conn.readHeader();
      while (conn.readLength == 64) {
        r = new Rule();
        r.fw_line = conn.in.readInt();
        r.level_num = conn.in.readInt();
        r.timeout = null;
        tmpint = conn.in.readInt();
        if (tmpint < 0)
          r.timeout_type = tmpint;
        else {
          r.timeout_type = Rule.SF_TIMEOUT_VALID;
          r.timeout = new Date((long)tmpint * 1000);
        }
        r.fw_rc = conn.in.readInt();
        r.fw_src_idx = conn.in.readInt();
        r.fw_src_cnt = conn.in.readInt();
        r.fw_dst_idx = conn.in.readInt();
        r.fw_dst_cnt = conn.in.readInt();
        r.fw_flags = conn.in.readInt();
        r.ttl = conn.in.readInt();
        r.protocol = conn.in.readInt();
        r.fw_rip_idx = conn.in.readInt();
        r.fw_rip_cnt = conn.in.readInt();
        r.usage = conn.in.readInt();
        r.bytes = (long)conn.in.readInt();
        tmpint = conn.in.readInt();
        r.bytes += (long)tmpint * Rule.GIGA;
        host.rules.addElement(r);
        conn.readHeader();
      }
      if (conn.readLength != 0) {
        commError = COMM_ERROR;
        return false;
      }

      conn.writeHeader(SERV_READ_VARS, 0);
      conn.readHeader();
      switch (conn.readCommand) {
        case SERV_OK:
          break;
        default:
          commError = COMM_ERROR;
          conn.in.skipBytes(conn.readLength);
          return false;
      } // switch
      // read number of variables
      conn.readHeader();
      if (conn.readLength != 4) {
        commError = COMM_ERROR;
        return false;
      }
      int numVars = conn.in.readInt();
      int varcnt = 0;
      host.varnames = new String[numVars];
      host.varvalues = new int[numVars][];
      host.varaddr = new byte[numVars][][];
      host.vartimeout = new Date[numVars][];
      conn.readHeader();
      byte buf[] = new byte[256];
      while (conn.readLength != 0) {
        // read variable name
        conn.in.readFully(buf, 0, conn.readLength);
        host.varnames[varcnt] = new String(buf, 0, conn.readLength);
        
        // read number of entries
        conn.readHeader();
        if (conn.readLength != 4) {
          commError = COMM_ERROR;
          return false;
        }
        numEntries = conn.in.readInt();
        host.varvalues[varcnt] = new int[numEntries];
        host.varaddr[varcnt] = new byte[numEntries][4];
        host.vartimeout[varcnt] = new Date[numEntries];

        for (int i=0; i < numEntries; i++) {
          conn.readHeader();
          if (conn.readLength != 12) {
            commError = COMM_ERROR;
            return false;
          }
          host.varvalues[varcnt][i] = conn.in.readInt();
          conn.in.readFully(host.varaddr[varcnt][i], 0, 4);
          tmpint = conn.in.readInt();
          host.vartimeout[varcnt][i] = new Date((long)tmpint * 1000);
        }

        conn.readHeader();
        varcnt++;
      }
      if (conn.readLength != 0) {
        commError = COMM_ERROR;
        return false;
      }

      return true; 
    }
    catch (Exception e) {
      dispose(conn);
      commError = COMM_IO;
      return false;
    } 
  }

  /**
   * Resolve host name
   * @param hname hostname or string containing IP address
   * @return IP address if successful, commError is set when returning null
   */

	public static synchronized byte[] resolve(String hname) {

		InetAddress iaddr;

		try {
			return Utils.parseIP(hname);
		}
		catch (IllegalArgumentException e) {}

		try {
			iaddr = InetAddress.getByName(hname);
			return iaddr.getAddress();
		}
		catch (UnknownHostException e) {
      commError = COMM_RESOLVE;
      return null;
		}
	} // resolve
			
  /**
   * Resolve IP address
   * @param addr IP address
   * @return hostname or IP address if not resolvable
   */
	public static synchronized String resolveIP(byte[] addr) {

		String addrString;

		if ((addr == null) || (addr.length != 4))
      return null;
    else {
      addrString = "" + Utils.unsign(addr[0]);
      for (int i=1; i < 4; i++)
        addrString += ("." + Utils.unsign(addr[i]));
		}

		try {
			InetAddress iaddr = InetAddress.getByName(addrString);
			return iaddr.getHostName();
		}
		catch (UnknownHostException e) {
			return addrString;
		}
	} // resolveIP

  /**
   * Resolve service number
   * @param port service number
   * @param proto protocol
   * @param privOnly resolve only if port < 1024
   * @return service name or number if not resolvable
   */
  public static synchronized String resolvePort(int port, 
    String proto, boolean privOnly) {
    if (port == 7227)
      return "firewall";
    String portString = Integer.toString(port);
    if (privOnly && port >= 1024)
      return portString;
    Hashtable portCache = proto.toUpperCase().equals("UDP") ? portCacheUDP : portCacheTCP;
    String portName = (String)portCache.get(portString);
    if (portName != null)
      return portName;
		// Port Name resolving should be done here!
		return "" + port;
	} // resolvePort
			

	/**
	 * Close connection to server.
	 * @param server Connection to server.
	 */
	public static synchronized void dispose(Connection conn) {

		UserDialog.ErrorBox("Connection to "+conn.iaddr.getHostName()+" lost!");
		conn.close();
		openConns.removeElement(conn);

	} // dispose
	
  /**
   * Close all connections
   */
  public static synchronized void dispose() {

	Connection conn;

    try {
      Enumeration servers = openConns.elements();
      while (servers.hasMoreElements()) {
				conn = (Connection)servers.nextElement();
				conn.close();
			}
			// re-initialize openConns
			openConns = new Vector(5, 5);
    }
    catch (Exception e) {}
  } // dispose


  // public data

  /**
   * Value of commError: successful operation
   */
  public final static int COMM_OK          =    0;
  /**
   * Value of commError: unspecified error
   */
  public final static int COMM_ERROR       =   -1;
  /**
   * Value of commError: cannot connect to firewall
   */
  public final static int COMM_NOCONN      =   -2;
  /**
   * Value of commError: general I/O error (e.g. connection aborted)
   */
  public final static int COMM_IO          =   -3; 
  /**
   * Value of commError: connect to server
   */
  public final static int COMM_OPENCONN    =   -4; 
  /**
   * Value of commError: error handling a local file
   */
  public final static int COMM_FILEERR     =   -5; 
  /**
   * Value of commError: error resolving hostname, IP address, service name
   * or port number
   */
  public final static int COMM_RESOLVE     =   -6; 
  /**
   * Value of commError: The parameter passed to the method does not refer
   * to a firewall.
   */
  public final static int COMM_NOFIREWALL  =   -7;
  /**
   * Value of commError: The host object passed to the method does not
   * contain an IP address.
   */
  public final static int COMM_NOADDRESS   =   -8;
  /**
   * Value of commError: The host object passed to the method does not
   * contain an IP address.
   */
  public final static int COMM_AUTH_FAILED   =   -9;

  /**
   * The variable commError holds the error code of the last operation
   */
  public static int commError = COMM_OK; // errorcode for last command

  /**
   * The variable parent holds the parent Frame
   */
  public static Frame parent; 				// parent frame

  /**
   * TCP port on which firewalls are accepting connections
   */
	protected final static int serverPort = 7227;

  // magic used in server communication
  private final static int magic = 0xbedd1bed;

  /**
   * Command used in communication with the server.
   */
  protected final static int SERV_SERVER_HELLO			= 0x0001;
  /**
   * Command used in communication with the server.
   */
  protected final static int SERV_CLIENT_HELLO				= 0x0002;
  /**
   * Command used in communication with the server.
   */
  protected final static int SERV_CLOSE							= 0x0003;
  /**
   * Command used in communication with the server.
   */
  protected final static int SERV_CLOSE_ACK					= 0x0004;
  /**
   * Command used in communication with the server.
   */
  protected final static int SERV_PING							= 0x0005;
  /**
   * Command used in communication with the server.
   */
  protected final static int SERV_PING_REPLY				= 0x0006;
  /**
   * Command used in communication with the server.
   */
  protected final static int SERV_GET_FILE					= 0x0007;
  /**
   * Command used in communication with the server.
   */
  protected final static int SERV_FILE_DATA					= 0x0008;
  /**
   * Command used in communication with the server.
   */
  protected final static int SERV_WRITE_FILE				= 0x0009;
  /**
   * Command used in communication with the server.
   */
  protected final static int SERV_WRITE_OK					= 0x000A;
  /**
   * Command used in communication with the server.
   */
  protected final static int SERV_WRITE_FAIL				= 0x000B;
  /**
   * Command used in communication with the server.
   */
  protected final static int SERV_READ_ADDR					= 0x000C;
  /**
   * Command used in communication with the server.
   */
  protected final static int SERV_READ_RULES				= 0x000D;
  /**
   * Command used in communication with the server.
   */
  protected final static int SERV_READ_VARS					= 0x000E;
  /**
   * Command used in communication with the server.
   */
  protected final static int SERV_RECONFIG					= 0x000F;
  /**
   * Command used in communication with the server.
   */
  protected final static int SERV_CHECKCONFIG				= 0x0010;
  /**
   * Command used in communication with the server.
   */
  protected final static int SERV_OK								= 0x0011;
  /**
   * Command used in communication with the server.
   */
  protected final static int SERV_ERROR							= 0x0012;
  /**
   * Command used in communication with the server.
   */
  protected final static int SERV_READ_TCP					= 0x0013;
  /**
   * Command used in communication with the server.
   */
  protected final static int SERV_KILL_TCP					= 0x0014;
  /**
   * Command used in communication with the server.
   */
  protected final static int SERV_RESET_USAGE				= 0x0015;
  /**
   * Command used in communication with the server.
   */
  protected final static int SERV_AUTHENTICATE			= 0x0016;
  /**
   * Command used in communication with the server.
   */
  protected final static int SERV_ERROR_AUTH				= 0x0017;
  /**
   * Command used in communication with the server.
   */
  protected final static int SERV_NOTAUTHENTICATED	= 0x0018;
  /**
   * Command used in communication with the server.
   */
  protected final static int SERV_SERVER_AUTHOK			= 0x0019;
  /**
   * Command used in communication with the server.
   */
  protected final static int SERV_CLIENT_AUTHOK			= 0x001A;
  /**
   * Command used in communication with the server.
   */
  protected final static int SERV_KEY_EXPIRED				= 0x001B;
  /**
   * Command used in communication with the server.
   */
  protected final static int SERV_KEY_UPDATE				= 0x001C;



  // private data
	private static String login;
	private static String key;
  private static Vector openConns = new Vector(5,5);  // open server connections
  private static Hashtable portCacheUDP = new Hashtable(); // Cache for UDP port translation
  private static Hashtable portCacheTCP = new Hashtable(); // Cache for TCP port translation
  private static Hashtable IPCache = new Hashtable();      // Cache for IP address translation

} // Communicator

